// Copyright (C) 1999 Dominic Letourneau and Andre Charbonneau
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this file.  If not, write to the Free Software Foundation,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/*
** Includes
*/
#include "FP_node_info_container.h"
#include "FileParserKeys.h"

// Output operators for parameter and inputs container classes...
ostream& operator<<(ostream& os, const param_info& pi)
{
  os << "<" << PARAMETER_KEY << ": " << pi.param_name << ", ";
  switch(pi.param_type)
    {
    case INT_PARAM:
      {
	os << INT_KEY << ":" << pi.param_value;
	break;
      }
    case SHORT_PARAM:
      {
	os << SHORT_KEY << ":" << pi.param_value;
	break;
      }
    case FLOAT_PARAM:
      {
	os << FLOAT_KEY << ":" << pi.param_value;
	break;
      }
    case DOUBLE_PARAM:
      {
	os << DOUBLE_KEY << ":" << pi.param_value;
	break;
      }
    case LONG_PARAM:
      {
	os << LONG_KEY << ":" << pi.param_value;
	break;
      }
    case UCHAR_PARAM:
      {
	os << UNSIGNED_CHAR_KEY << ":" << pi.param_value;
	break;
      }
    case UINT_PARAM:
      {
	os << UNSIGNED_INT_KEY << ":" << pi.param_value;
	break;
      }
    case ULONG_PARAM:
      {
	os << UNSIGNED_LONG_KEY << ":" << pi.param_value;
	break;
      }
    case STRING_PARAM:
      {
	os << STRING_KEY << ":" << "\"" << pi.param_value << "\"";
	break;
      }
    case BOOL_PARAM:
      {
	os << BOOL_KEY << ":" << pi.param_value;
	break;
      }
    case CHAR_PARAM:
      {
	os << CHAR_KEY << ":'" << pi.param_value << "'";
	break;
      }
    case SUBNET_PARAM:
      {
	os << SUBNET_PARAMETER_KEY << ":" << pi.param_value;
	break;
      }
    }
  os << ">";

  return os;
}

ostream& operator<<(ostream& os, const input_var_info& ii)
{
  os << "<input: " << ii.var_name << ", " << ii.source_node_name << ", " << ii.source_var_name << ">";
  return os;
}


FP_node_info_container::FP_node_info_container()
{
  // nothing to do ...
  return;
}

FP_node_info_container::FP_node_info_container(const FP_node_info_container& nic)
{
  m_node_name = nic.m_node_name;
  m_node_type = nic.m_node_type;
  m_params_map = nic.m_params_map;
  m_inputs_map = nic.m_inputs_map;
  return;
}

FP_node_info_container::~FP_node_info_container()
{
  // nothing to do...
}

FP_node_info_container& FP_node_info_container::operator=(const FP_node_info_container& nic)
{
  m_node_name = nic.m_node_name;
  m_node_type = nic.m_node_type;
  m_params_map = nic.m_params_map;
  m_inputs_map = nic.m_inputs_map;
  return *this;
}

void FP_node_info_container::set_node_name(const string& node_name)
{
  m_node_name = node_name;
  return;
}

string FP_node_info_container::get_node_name() const
{
  return m_node_name;
}

void FP_node_info_container::set_node_type(const string& node_type)
{
  m_node_type = node_type;
  return;
}

string FP_node_info_container::get_node_type() const
{
  return m_node_type;
}


void FP_node_info_container::add_input_var(const string& var_name, const string& source_node_name, const string& source_var_name) throw(FileParserException)
{
  if(m_inputs_map.count(var_name) > 0)
    {
      throw(FileParserException(DUPLICATE_EXCEPTION,string("Tried to add input variable ") + var_name + string(" twice.")));
    }
  else
    {
      input_var_info info;
      info.var_name = var_name;
      info.source_node_name = source_node_name;
      info.source_var_name = source_var_name;
      m_inputs_map[var_name] = info;
    }

  return;
}

input_var_info FP_node_info_container::get_input_var_info(const string& input_name) const throw(FileParserException)
{
  input_var_info return_value;
  if(m_inputs_map.count(input_name) == 0)
    {
      throw(FileParserException(NOT_FOUND_EXCEPTION,string("Input variable ") + input_name + string(" not found.")));
    }
  else
    {
      return_value = m_inputs_map.find(input_name)->second;
    }
  
  return return_value;
}

void FP_node_info_container::add_param_var(const string& param_name, const string& param_value, const int param_type) throw(FileParserException)
{
  if(m_params_map.count(param_name) > 0)
    {
      throw(FileParserException(DUPLICATE_EXCEPTION,string("Tried to add parameter variable ") + param_name + string(" twice.")));
    }
  else
    {
      param_info pi;
      pi.param_name = param_name;
      pi.param_value = param_value;
      pi.param_type = param_type;
      m_params_map[param_name] = pi;
    }

  return;
}


void FP_node_info_container::modify_param_var(const string& param_name, const string& param_value, const int param_type) throw(FileParserException)
{
  if(m_params_map.count(param_name) == 0)
    {
      throw(FileParserException(DUPLICATE_EXCEPTION,string("Tried to modify non-existing parameter variable ") + param_name));
    }
  else
    {
      param_info pi;
      pi.param_name = param_name;
      pi.param_value = param_value;
      pi.param_type = param_type;
      m_params_map[param_name] = pi;
    }

  return;  
}


param_info FP_node_info_container::get_param_info(const string& param_name) const throw(FileParserException)
{
  param_info return_value;
  if(m_params_map.count(param_name) == 0)
    {
      throw(FileParserException(NOT_FOUND_EXCEPTION,string("Parameter variable ") + param_name + string(" not found.")));
    }
  else
    {
      return_value = m_params_map.find(param_name)->second;
    }
  
  return return_value;
}

list<string> FP_node_info_container::get_input_var_names() const
{
  list<string> return_value;
  
  for(map<string,input_var_info>::const_iterator i = m_inputs_map.begin(); i != m_inputs_map.end(); i++)
    {
      return_value.push_back(i->first);
    }

  return return_value;
}

list<string> FP_node_info_container::get_param_var_names() const
{
  list<string> return_value;

  for(map<string,param_info>::const_iterator i = m_params_map.begin(); i != m_params_map.end(); i++)
    {
      return_value.push_back(i->first);
    }

  return return_value;
}

void FP_node_info_container::reset_params()
{
  m_params_map.erase(m_params_map.begin(), m_params_map.end());
  return;
}

ostream& operator<<(ostream& os, const FP_node_info_container& nic)
{
  os << "<node: " << nic.get_node_name() << ">" << "\t" << "<type: " << nic.get_node_type() << ">";

  for(map<string,param_info>::const_iterator i = nic.m_params_map.begin();
      !(i == nic.m_params_map.end());
      i++)
    {
      os << endl << "\t" << i->second;
    }

  for(map<string,input_var_info>::const_iterator j = nic.m_inputs_map.begin();
      !(j == nic.m_inputs_map.end());
      j++)
    {
      os << endl << "\t" << j->second;
    }

  return os;
}
