// Copyright (C) 1999  Dominic Letourneau and Andre Charbonneau
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this file.  If not, write to the Free Software Foundation,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "NetworkBuilder.h"
#include "Iterator.h"
#include <stdio.h>
#include <algorithm>
#include <string>
#include <list>

list<string> NetworkBuilder::create_building_list(const map<string,FP_network_info_container>& net_info_map, const string& net_name) const throw(NetworkBuilderException)
{
  list<string> building_list;
  list<string> nodes_list;
  list<string> nets_to_check;
  list<string> recursion_checking_list;

  // First lets check if net_name is in the network info map.
  if(net_info_map.count(net_name) == 0)
    {
      throw(NetworkBuilderException(NB_NOT_FOUND_EXCEPTION,string("Network named: ") + net_name + string(" was not found in network info map.")));
    }


  nets_to_check.push_front(net_name);
  recursion_checking_list.push_front(net_name);

  while(!nets_to_check.empty())
    {
      list<string> nodes_type_list;

      string net_to_check = nets_to_check.front();
      nets_to_check.pop_front();

      // add it to the building list...
      building_list.push_front(net_to_check);
      
      // get its nodes type list
      nodes_list = net_info_map.find(net_to_check)->second.get_node_names();
      for(list<string>::const_iterator i = nodes_list.begin(); !(i == nodes_list.end()); i++)
	{
	  nodes_type_list.push_back(net_info_map.find(net_to_check)->second.get_type_of_node(*i));
	}

      nodes_type_list.unique();
      
      // now for each node type, check if it is a net...
      for(list<string>::const_iterator j = nodes_type_list.begin(); !(j == nodes_type_list.end()); j++)
	{
	  if(net_info_map.count(*j) > 0)
	    {
	      // its a subnet, add it to the list...
	      nets_to_check.push_front(*j);
	      recursion_checking_list.push_front(*j);
	    }
	}

      // check for recursion
      int size_before_unique = recursion_checking_list.size();
      recursion_checking_list.sort();
      recursion_checking_list.unique();
      int size_after_unique = recursion_checking_list.size();
      if(size_before_unique != size_after_unique)
	{
	  throw(NetworkBuilderException(NB_RECURSION_EXCEPTION,"Recursivity in subnets found."));
	}
    }

  if(verbose())
    {
      cout << "Order of network building process:" << endl;
      cout << "----------------------------------" << endl;
      for(list<string>::const_iterator i = building_list.begin();
	  !(i == building_list.end());
	  i++)
	{
	  cout << "\t" << *i << endl;
	}
      cout << "----------------------------------" << endl << endl;
    }
  

  return building_list;
}

map<string,FP_network_info_container> NetworkBuilder::crunch_subnets(const map<string,FP_network_info_container>& net_info_map, const string& net_name) throw(NetworkBuilderException)
{
  bool done = false;
  bool found_subnet = false;
  map<string,FP_network_info_container> return_value = net_info_map;
  

  try
    {

      /*
      ** Start of recursion checking...
      */
      list<string> nodes_list;
      list<string> nets_to_check;
      list<string> recursion_checking_list;
      
      // First lets check if net_name is in the network info map.
      if(net_info_map.count(net_name) == 0)
	{
	  throw(NetworkBuilderException(NB_NOT_FOUND_EXCEPTION,string("Network named: ") + net_name + string(" was not found in network info map.")));
	}
      
      
      nets_to_check.push_front(net_name);
      recursion_checking_list.push_front(net_name);
      
      while(!nets_to_check.empty())
	{
	  list<string> nodes_type_list;
	  
	  string net_to_check = nets_to_check.front();
	  nets_to_check.pop_front();
	  
	  // get its nodes type list
	  nodes_list = net_info_map.find(net_to_check)->second.get_node_names();
	  for(list<string>::const_iterator i = nodes_list.begin(); !(i == nodes_list.end()); i++)
	    {
	      nodes_type_list.push_back(net_info_map.find(net_to_check)->second.get_type_of_node(*i));
	    }
	  nodes_type_list.sort();
	  nodes_type_list.unique();
	  
	  // now for each node type, check if it is a net...
	  for(list<string>::const_iterator j = nodes_type_list.begin(); !(j == nodes_type_list.end()); j++)
	    {
	      if(net_info_map.count(*j) > 0)
		{
		  // its a subnet, add it to the list...
		  nets_to_check.push_front(*j);
		  recursion_checking_list.push_front(*j);
		}
	    }
	  
	  // check for recursion
	  recursion_checking_list.sort();
	  if(adjacent_find(recursion_checking_list.begin(), recursion_checking_list.end()) != recursion_checking_list.end())
	    {
	      throw(NetworkBuilderException(NB_RECURSION_EXCEPTION,string("Recursivity in network ") + *(adjacent_find(recursion_checking_list.begin(), recursion_checking_list.end()))));
	    }
	}
      /*
      ** End of recursion checking...
      */
      
      




      while(!done)
	{
	  found_subnet = false;
	  list<FP_network_info_container> new_subnets;
	  
	  for(map<string,FP_network_info_container>::iterator net_iter = return_value.begin(); 
	      net_iter != return_value.end(); 
	      net_iter++)
	    {
	      
	      list<string> nodes_list;
	      
	      // get its nodes
	      nodes_list = net_iter->second.get_node_names();
	      
	      // loop through its nodes...
	      for(list<string>::iterator node_name_iter = nodes_list.begin();
		  node_name_iter != nodes_list.end();
		  node_name_iter++)
		{
		  // Check to see if the node is a subnet...
		  if(net_info_map.count(net_iter->second.get_type_of_node(*node_name_iter)) != 0)
		    {
		      // --- its a subnet ---
		      found_subnet = true;
		      
		      // find a unique name for the subnet...
		      string unique_net_name = 
			find_unique_net_name(net_info_map, net_iter->second.get_type_of_node(*node_name_iter));
		      
		      // create a copy of the subnet in question
		      FP_network_info_container new_net_info_container = 
			net_info_map.find(net_iter->second.get_type_of_node(*node_name_iter))->second;
		      
		      // set new type and subnet name...
		      new_net_info_container.set_name(unique_net_name);
		      net_iter->second.set_type_of_node(*node_name_iter,unique_net_name);
		      
		      
		      // get a list of param_info structs from the wrapping node...
		      list<param_info> wrapping_node_pi_list = net_iter->second.get_params_from_node(*node_name_iter);
		      
		      // get the list of node names from the subnet
		      list<string> new_subnet_node_names = new_net_info_container.get_node_names();
		      
		      // now go throught each node of the new subnet and modify its parameters accordingly
		      for(list<string>::iterator j = new_subnet_node_names.begin();
			  j != new_subnet_node_names.end();
			  j++)
			{
			  list<param_info> subnet_node_pi_list = new_net_info_container.get_params_from_node(*j);
			  // loop through each parameter...
			  for(list<param_info>::iterator k = subnet_node_pi_list.begin();
			      k != subnet_node_pi_list.end();
			      k++)
			    {
			      if(k->param_type == SUBNET_PARAM)
				{
				  string name_to_look_for = k->param_value;
				  for(list<param_info>::iterator m = wrapping_node_pi_list.begin();
				      m != wrapping_node_pi_list.end();
				      m++)
				    {
				      if(m->param_name == name_to_look_for)
					{
					  param_info new_pi = *k;
					  new_pi.param_type = m->param_type;
					  new_pi.param_value = m->param_value;
					  new_net_info_container.modify_param_to_node(*j,new_pi);
					}
				    }
				}
			    }
			}
		      
		      // remove subnet parameters from wrapping node
		      net_iter->second.reset_params_from_node(*node_name_iter);
		      
		      // add it to the new subnet list...
		      new_subnets.push_back(new_net_info_container);
		    }
		}
	    }
	  
	  for(list<FP_network_info_container>::iterator i = new_subnets.begin();
	      !(i == new_subnets.end());
	      i++)
	    {
	      return_value[i->get_name()] = (*i);
	    }
	  
	  if(!found_subnet)
	    {
	      done = true;
	    }
	}
    }
  catch(FileParserException& e)
    {
      e.print();
      throw(NetworkBuilderException(NB_NETWORK_PARSER_EXCEPTION,"FileParserException caught."));
    }
  catch(NetworkBuilderException)
    {
      throw;
    }


  if(verbose())
    {
      cout << "---------- After subnet crunching: ----------" << endl;
      for(map<string,FP_network_info_container>::iterator i = return_value.begin();
	  !(i == return_value.end());
	  i++)
	{
	  cout << i->second << endl;
	}
    }

  return return_value;
}

string NetworkBuilder::find_unique_net_name(const map<string,FP_network_info_container>& net_info_map, const string& prefix)
{
  unsigned int unique_id_number = 0;
  string unique_name;
  static list<string> unique_names_generated;
  //bool b;
  do
    {
      char buf[10];
      sprintf(buf,"%i",unique_id_number);
      unique_id_number++;
      unique_name = prefix + string("__") + string(buf);
    }
  while((net_info_map.count(unique_name) != 0) || 
	!(find(unique_names_generated.begin(), unique_names_generated.end(), unique_name) == unique_names_generated.end()));

  unique_names_generated.push_back(unique_name);
  return unique_name;
}

NetworkBuilder::NetworkBuilder()
{
  m_verbose = false;
  return;
}

NetworkBuilder::~NetworkBuilder()
{
  return;
}

Network* NetworkBuilder::build(const map<string,FP_network_info_container>& nim, const string& net_name, ParameterSet* external_ps) throw(NetworkBuilderException)
{
  list<Network*> nets_created;
  list<string> building_list;
  map<string,FP_network_info_container> net_info_map = nim;

  // First lets check if net_name is in the network info map.
  if(nim.count(net_name) == 0)
    {
      throw(NetworkBuilderException(NB_NOT_FOUND_EXCEPTION,string("Network named: ") + net_name + string(" was not found in network info map.")));
    }

  try
    {
      // start by crunching the subnets...
      
      net_info_map = crunch_subnets(net_info_map, net_name);

      building_list = create_building_list(net_info_map, net_name);

      for(list<string>::const_iterator net_iter = building_list.begin(); !(net_iter == building_list.end()); net_iter++)
	{
	  Network* net_ptr = NULL;
	  if(net_info_map.find(*net_iter)->second.is_iterator())
	    {
	      if(verbose())
		{
		  cout << "NetworkBuilder -> Building iterator data-flow network:" << *net_iter << "..." << endl;
		}
	      net_ptr = new Iterator(*net_iter,ParameterSet());
	    }
	  else
	    {
	      if(verbose())
		{
		  cout << "NetworkBuilder -> Building data-flow network:" << *net_iter << "..." << endl;
		}
	      net_ptr = new Network(*net_iter,ParameterSet());
	    }

	  list<string> node_names = net_info_map.find(*net_iter)->second.get_node_names();

	  for(list<string>::const_iterator node_iter = node_names.begin(); !(node_iter == node_names.end()); node_iter++)
	    {
	      // first check if the node is a subnet already created...
	      bool is_subnet = false;
	      list<Network*>::const_iterator subnet_iter;
	      for(subnet_iter = nets_created.begin();
		  !(subnet_iter == nets_created.end());
		  subnet_iter++)
		{
		  if((*subnet_iter)->getName() == net_info_map.find(*net_iter)->second.get_type_of_node(*node_iter))
		    {
		      is_subnet = true;
		      break;
		    }
		}

	      if(is_subnet)
		{
		  // add the subnet node
		  (*(*subnet_iter)).setName(*node_iter);
		  if(verbose())
		    {
		      cout << " **** Adding subnet: " << (*(*subnet_iter)).getName() << endl;
		    }
		  net_ptr->addNode((*(*subnet_iter)));
		}
	      else
		{
		  list<string> param_var_names;
		  // Get the parameters of the node
		  ParameterSet ps;
		  param_var_names = net_info_map.find(*net_iter)->second.get_param_var_names_from_node(*node_iter);
		  for(list<string>::const_iterator param_iter = param_var_names.begin(); 
		      !(param_iter == param_var_names.end()); param_iter++)
		    {
		      param_info pi = net_info_map.find(*net_iter)->second.get_param_from_node(*node_iter,*param_iter);
		      switch(pi.param_type)
			{
			case INT_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new Int(atoi(pi.param_value.c_str()))));
			    break;
			  }
			case SHORT_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new Short(atoi(pi.param_value.c_str()))));
			    break;
			  }
			case FLOAT_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new Float(atof(pi.param_value.c_str()))));
			    break;
			  }
			case DOUBLE_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new Double(atof(pi.param_value.c_str()))));
			    break;
			  }
			case LONG_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new Long(atol(pi.param_value.c_str()))));
			    break;
			  }
			case UCHAR_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new U_char(atoi(pi.param_value.c_str()))));
			    break;
			  }
			case UINT_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new U_int(atoi(pi.param_value.c_str()))));
			    break;
			  }
			case ULONG_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new U_int(atol(pi.param_value.c_str()))));
			    break;
			  }
			case STRING_PARAM:
			  {
			    ps.add(pi.param_name,ObjectRef(new String(pi.param_value)));
			    break;
			  }
			case BOOL_PARAM:
			  {
			    if(pi.param_value == "true")
			      {
				ps.add(pi.param_name,ObjectRef(new Bool(true)));
			      }
			    else
			      {
				ps.add(pi.param_name,ObjectRef(new Bool(false)));
			      }
			    break;
			  }
			case EXT_PARAM:
			  {
			    if(external_ps == NULL)
			      {
				throw(NetworkBuilderException(NB_EXTERNAL_PARAMETER_EXCEPTION, 
							      "No external parameter set given to network builder."));
			      }
			    else
			      {
				if(!external_ps->exist(pi.param_value))
				  {
				    throw(NetworkBuilderException(NB_EXTERNAL_PARAMETER_EXCEPTION, 
								  string("External variable: ") + pi.param_value + string(" not found in external parameter set.")));
				  }
				else
				  {
				    ps.add(pi.param_name,external_ps->get(pi.param_value));
				  }

			      }
			    break;
			  }
			default:
			  {
			    throw(NetworkBuilderException(NB_FATAL_ERROR,
							  "Parameter type not recognized by the network builder."));
			    break;
			  }
			}
		    }
		  // add the node
		  if(verbose())
		    {
		      cout << "**** Adding node: " << "\t[" << *node_iter << "]\tto:\t[" << *net_iter << "]" << endl;
		    }
		  net_ptr->addNode(net_info_map.find(*net_iter)->second.get_type_of_node(*node_iter),*node_iter,ps);
		}
	    } // end of node iteratrion

	  // Connect the nodes together...
	  
	  for(list<string>::const_iterator node_iter2 = node_names.begin(); !(node_iter2 == node_names.end()); node_iter2++)
	    {
	      list<string> input_var_names;
	      input_var_names = net_info_map.find(*net_iter)->second.get_input_var_names_from_node(*node_iter2);
	      for(list<string>::const_iterator input_iter = input_var_names.begin(); 
		  !(input_iter == input_var_names.end()); input_iter++)
		{
		  input_var_info ii = net_info_map.find(*net_iter)->second.get_input_from_node(*node_iter2,*input_iter);
		  if(verbose())
		    {
		      cout << "\tConnecting node:" << *node_iter2 << "\tto node:" << ii.source_node_name << 
			"\t[" << ii.var_name << " <-> " << ii.source_var_name << "]" << endl;
		    }
		  net_ptr->connect(*node_iter2, ii.var_name, ii.source_node_name, ii.source_var_name);
		}
	    }


	  // Set sink and source node...
	  if(net_info_map.find(*net_iter)->second.get_input_node_name() != "")
	    {
	      if(verbose())
		{
		  cout << "Setting input node of network: " << *net_iter 
		       << " to: " << net_info_map.find(*net_iter)->second.get_input_node_name() << endl;
		}
	      net_ptr->setInputNode(net_ptr->getNodeNamed(net_info_map.find(*net_iter)->second.get_input_node_name()));
	    }
	  if(net_info_map.find(*net_iter)->second.get_output_node_name() != "")
	    {
	      if(verbose())
		{
		  cout << "Setting output node of network: " << *net_iter 
		       << " to: " << net_info_map.find(*net_iter)->second.get_output_node_name() << endl;
		}
	      net_ptr->setSinkNode(net_ptr->getNodeNamed(net_info_map.find(*net_iter)->second.get_output_node_name()));
	    }


	  // Set condition node (for iterator nets only...)
	  if(net_info_map.find(*net_iter)->second.is_iterator() && net_info_map.find(*net_iter)->second.get_condition_node_name() == "")
	    {
	      throw(NetworkBuilderException(NB_ITERATOR_NET_EXCEPTION, string("Condition node not present in iterator net: ") + *net_iter));
	    }
	  if(!(net_info_map.find(*net_iter)->second.is_iterator()) && net_info_map.find(*net_iter)->second.get_condition_node_name() != "")
	    {
	      throw(NetworkBuilderException(NB_ITERATOR_NET_EXCEPTION, string("Condition node found in non iterator net: ") + *net_iter));
	    }
	  
	  if(net_info_map.find(*net_iter)->second.is_iterator())
	    {
	      if(verbose())
		{
		  cout << "Setting iterator node of network: " << *net_iter << " to: " 
		       << net_info_map.find(*net_iter)->second.get_condition_node_name() << endl;
		}
	      ((Iterator*)net_ptr)->setConditionNode(net_ptr->getNodeNamed(net_info_map.find(*net_iter)->second.get_condition_node_name()));
	    }
	  
	  // Add the net to the net list...
	  nets_created.push_back(net_ptr);

	} // end of net iteration
    }
  catch(NetworkBuilderException)
    {
      throw;
    }
  catch(BaseException &e)
    {
      e.print(cerr);
      throw(NetworkBuilderException(NB_NETWORK_EXCEPTION,"BaseException caught during network building."));
    }
  catch(...)
    {
      throw(NetworkBuilderException(NB_FATAL_ERROR,"Unhandeled exception caught during network building."));
    }


  return nets_created.back();
}


void NetworkBuilder::set_verbose(const bool b)
{
  m_verbose = b;
  return;
}


bool NetworkBuilder::verbose() const
{
  return m_verbose;
}
